
from dynamof.operations.add import add
from dynamof.operations.create import create
from dynamof.operations.delete import delete
from dynamof.operations.describe import describe
from dynamof.operations.find import find
from dynamof.operations.query import query
from dynamof.operations.update import update
from dynamof.operations.scan import scan

from dynamof.attribute import attr, cand, cor

from dynamof.db import db
from dynamof.table import table




from dynamof.core.utils import merge, find, flatten
from dynamof.core.model import Attr, Condition, Function
from dynamof.core.dynamo import get_safe_alias


# condition and
def cand(*conditions):
    def build(attrs):
        return ' AND '.join([f'({cond.expression(attrs)})' for cond in conditions])
    attributes = merge([cond.attributes for cond in conditions])
    references = flatten([cond.references for cond in conditions])
    return Condition(build, attributes, references=references)

# condition or
def cor(*conditions):
    def build(attrs):
        return ' OR '.join([f'({cond.expression(attrs)})' for cond in conditions])
    attributes = merge([cond.attributes for cond in conditions])
    references = flatten([cond.references for cond in conditions])
    return Condition(build, attributes, references=references)


class AttrFunc:

    def __init__(self):
        pass

    def __call__(self, name):

        def find_attr(attrs, original_name=None):
            original_name = original_name or name
            return find(attrs, lambda a: a.get('original') == original_name)

        def equals(value):
            def build(attrs):
                a = find_attr(attrs)
                return f'{a.alias} = {a.key}'
            return Condition(build, { name: value })

        def greater_than(value):
            def build(attrs):
                a = find_attr(attrs)
                return f'{a.alias} > {a.key}'
            return Condition(build, { name: value })

        def less_than(value):
            def build(attrs):
                a = find_attr(attrs)
                return f'{a.alias} < {a.key}'
            return Condition(build, { name: value })

        def less_than_or_equal(value):
            def build(attrs):
                a = find_attr(attrs)
                return f'{a.alias} <= {a.key}'
            return Condition(build, { name: value })

        def greater_than_or_equal(value):
            def build(attrs):
                a = find_attr(attrs)
                return f'{a.alias} >= {a.key}'
            return Condition(build, { name: value })

        def between(value_a, value_b):
            value_a_key = f'{name}_a'
            value_b_key = f'{name}_b'
            def build(attrs):
                alias = get_safe_alias(name)
                value_a_attr = find_attr(attrs, original_name=value_a_key)
                value_b_attr = find_attr(attrs, original_name=value_b_key)
                return f'{alias} BETWEEN {value_a_attr.key} AND {value_b_attr.key}'
            return Condition(build, {
                value_a_key: value_a,
                value_b_key: value_b
            }, references=[name])

        def begins_with(value):
            def build(attrs):
                a = find_attr(attrs)
                return f'begins_with({a.alias}, {a.key})'
            return Condition(build, { name: value })

        return Attr(equals, greater_than, less_than, less_than_or_equal, greater_than_or_equal, between, begins_with)

    @classmethod
    def append(cls, *values):
        """Generates a function that will tell the builder the correct format
        for list appending for a specific item. The full update expression would
        look like...
            UpdateExpression="SET some_attr = list_append(some_attr, :i)"
        """
        def expression(a_attr):
            return f'{a_attr.alias} = list_append({a_attr.alias}, {a_attr.key})'

        def value():
            return list(values)

        return Function(expression, value)

    @classmethod
    def prepend(cls, *values):
        """Same as append - in reverse"""
        def expression(a_attr):
            return f'{a_attr.alias} = list_append({a_attr.key}, {a_attr.alias})'

        def value():
            return list(values)

        return Function(expression, value)


# NOTE: Magic here!!!
# Because Attr implments __call__
# this makes it a special object data type that
# can both have properties and be called:
#
#   roles: attr.append('admin')
#
# and
#
#   condition: attr('sunshie').equals('sunshie')
#

attr = AttrFunc()




from dynamof.core import builder as ab
from dynamof.core import args
from dynamof.core.utils import shake
from dynamof.core.model import Operation
from dynamof.core.response import response


def find(table_name: str, key: dict):
    """Creates an Operation that will find an item in a table when run.

    Args:
        table_name (str): The name of the table to find the item in
        key (dict): An object with a key, value that will be used as the identifier for the item
            Examples:
                - { 'username': 'sunshie' }
                - { 'id': '9snw82h', 'type': 'car' } -> when table uses hash key and range key

    """
    build = ab.builder(
        table_name=table_name,
        key=key)
    description = shake(
        TableName=build(args.TableName),
        Key=build(args.Key))
    return Operation(description, run)

def run(client, description):
    res = client.get_item(**description)
    return response(res)




from dynamof.core import builder as ab
from dynamof.core import args
from dynamof.core.utils import shake
from dynamof.core.model import Operation
from dynamof.core.response import response


def add(table_name: str, item: dict, auto_id: bool = None):
    """Creates an Operation that will add an item to a table when run.

    Args:
        table_name (str): The name of the table to add the item to.
        item (dict): The item to add to the table.

    """
    build = ab.builder(
        table_name=table_name,
        attributes=item,
        auto_id=auto_id)
    description = shake(
        TableName=build(args.TableName),
        Item=build(args.Item),
        ReturnValues='ALL_OLD')
    return Operation(description, run)

def run(client, description):
    res = client.put_item(**description)
    return response(res)





def get_safe_alias(attr_name):
    return DYNAMO_RESERVED_WORDS.get(attr_name.upper(), attr_name)

DYNAMO_RESERVED_WORDS = {
    'ABORT': '#abort',
    'ABSOLUTE': '#absolute',
    'ACTION': '#action',
    'ADD': '#add',
    'AFTER': '#after',
    'AGENT': '#agent',
    'AGGREGATE': '#aggregate',
    'ALL': '#all',
    'ALLOCATE': '#allocate',
    'ALTER': '#alter',
    'ANALYZE': '#analyze',
    'AND': '#and',
    'ANY': '#any',
    'ARCHIVE': '#archive',
    'ARE': '#are',
    'ARRAY': '#array',
    'AS': '#as',
    'ASC': '#asc',
    'ASCII': '#ascii',
    'ASENSITIVE': '#asensitive',
    'ASSERTION': '#assertion',
    'ASYMMETRIC': '#asymmetric',
    'AT': '#at',
    'ATOMIC': '#atomic',
    'ATTACH': '#attach',
    'ATTRIBUTE': '#attribute',
    'AUTH': '#auth',
    'AUTHORIZATION': '#authorization',
    'AUTHORIZE': '#authorize',
    'AUTO': '#auto',
    'AVG': '#avg',
    'BACK': '#back',
    'BACKUP': '#backup',
    'BASE': '#base',
    'BATCH': '#batch',
    'BEFORE': '#before',
    'BEGIN': '#begin',
    'BETWEEN': '#between',
    'BIGINT': '#bigint',
    'BINARY': '#binary',
    'BIT': '#bit',
    'BLOB': '#blob',
    'BLOCK': '#block',
    'BOOLEAN': '#boolean',
    'BOTH': '#both',
    'BREADTH': '#breadth',
    'BUCKET': '#bucket',
    'BULK': '#bulk',
    'BY': '#by',
    'BYTE': '#byte',
    'CALL': '#call',
    'CALLED': '#called',
    'CALLING': '#calling',
    'CAPACITY': '#capacity',
    'CASCADE': '#cascade',
    'CASCADED': '#cascaded',
    'CASE': '#case',
    'CAST': '#cast',
    'CATALOG': '#catalog',
    'CHAR': '#char',
    'CHARACTER': '#character',
    'CHECK': '#check',
    'CLASS': '#class',
    'CLOB': '#clob',
    'CLOSE': '#close',
    'CLUSTER': '#cluster',
    'CLUSTERED': '#clustered',
    'CLUSTERING': '#clustering',
    'CLUSTERS': '#clusters',
    'COALESCE': '#coalesce',
    'COLLATE': '#collate',
    'COLLATION': '#collation',
    'COLLECTION': '#collection',
    'COLUMN': '#column',
    'COLUMNS': '#columns',
    'COMBINE': '#combine',
    'COMMENT': '#comment',
    'COMMIT': '#commit',
    'COMPACT': '#compact',
    'COMPILE': '#compile',
    'COMPRESS': '#compress',
    'CONDITION': '#condition',
    'CONFLICT': '#conflict',
    'CONNECT': '#connect',
    'CONNECTION': '#connection',
    'CONSISTENCY': '#consistency',
    'CONSISTENT': '#consistent',
    'CONSTRAINT': '#constraint',
    'CONSTRAINTS': '#constraints',
    'CONSTRUCTOR': '#constructor',
    'CONSUMED': '#consumed',
    'CONTINUE': '#continue',
    'CONVERT': '#convert',
    'COPY': '#copy',
    'CORRESPONDING': '#corresponding',
    'COUNT': '#count',
    'COUNTER': '#counter',
    'CREATE': '#create',
    'CROSS': '#cross',
    'CUBE': '#cube',
    'CURRENT': '#current',
    'CURSOR': '#cursor',
    'CYCLE': '#cycle',
    'DATA': '#data',
    'DATABASE': '#database',
    'DATE': '#date',
    'DATETIME': '#datetime',
    'DAY': '#day',
    'DEALLOCATE': '#deallocate',
    'DEC': '#dec',
    'DECIMAL': '#decimal',
    'DECLARE': '#declare',
    'DEFAULT': '#default',
    'DEFERRABLE': '#deferrable',
    'DEFERRED': '#deferred',
    'DEFINE': '#define',
    'DEFINED': '#defined',
    'DEFINITION': '#definition',
    'DELETE': '#delete',
    'DELIMITED': '#delimited',
    'DEPTH': '#depth',
    'DEREF': '#deref',
    'DESC': '#desc',
    'DESCRIBE': '#describe',
    'DESCRIPTOR': '#descriptor',
    'DETACH': '#detach',
    'DETERMINISTIC': '#deterministic',
    'DIAGNOSTICS': '#diagnostics',
    'DIRECTORIES': '#directories',
    'DISABLE': '#disable',
    'DISCONNECT': '#disconnect',
    'DISTINCT': '#distinct',
    'DISTRIBUTE': '#distribute',
    'DO': '#do',
    'DOMAIN': '#domain',
    'DOUBLE': '#double',
    'DROP': '#drop',
    'DUMP': '#dump',
    'DURATION': '#duration',
    'DYNAMIC': '#dynamic',
    'EACH': '#each',
    'ELEMENT': '#element',
    'ELSE': '#else',
    'ELSEIF': '#elseif',
    'EMPTY': '#empty',
    'ENABLE': '#enable',
    'END': '#end',
    'EQUAL': '#equal',
    'EQUALS': '#equals',
    'ERROR': '#error',
    'ESCAPE': '#escape',
    'ESCAPED': '#escaped',
    'EVAL': '#eval',
    'EVALUATE': '#evaluate',
    'EXCEEDED': '#exceeded',
    'EXCEPT': '#except',
    'EXCEPTION': '#exception',
    'EXCEPTIONS': '#exceptions',
    'EXCLUSIVE': '#exclusive',
    'EXEC': '#exec',
    'EXECUTE': '#execute',
    'EXISTS': '#exists',
    'EXIT': '#exit',
    'EXPLAIN': '#explain',
    'EXPLODE': '#explode',
    'EXPORT': '#export',
    'EXPRESSION': '#expression',
    'EXTENDED': '#extended',
    'EXTERNAL': '#external',
    'EXTRACT': '#extract',
    'FAIL': '#fail',
    'FALSE': '#false',
    'FAMILY': '#family',
    'FETCH': '#fetch',
    'FIELDS': '#fields',
    'FILE': '#file',
    'FILTER': '#filter',
    'FILTERING': '#filtering',
    'FINAL': '#final',
    'FINISH': '#finish',
    'FIRST': '#first',
    'FIXED': '#fixed',
    'FLATTERN': '#flattern',
    'FLOAT': '#float',
    'FOR': '#for',
    'FORCE': '#force',
    'FOREIGN': '#foreign',
    'FORMAT': '#format',
    'FORWARD': '#forward',
    'FOUND': '#found',
    'FREE': '#free',
    'FROM': '#from',
    'FULL': '#full',
    'FUNCTION': '#function',
    'FUNCTIONS': '#functions',
    'GENERAL': '#general',
    'GENERATE': '#generate',
    'GET': '#get',
    'GLOB': '#glob',
    'GLOBAL': '#global',
    'GO': '#go',
    'GOTO': '#goto',
    'GRANT': '#grant',
    'GREATER': '#greater',
    'GROUP': '#group',
    'GROUPING': '#grouping',
    'HANDLER': '#handler',
    'HASH': '#hash',
    'HAVE': '#have',
    'HAVING': '#having',
    'HEAP': '#heap',
    'HIDDEN': '#hidden',
    'HOLD': '#hold',
    'HOUR': '#hour',
    'IDENTIFIED': '#identified',
    'IDENTITY': '#identity',
    'IF': '#if',
    'IGNORE': '#ignore',
    'IMMEDIATE': '#immediate',
    'IMPORT': '#import',
    'IN': '#in',
    'INCLUDING': '#including',
    'INCLUSIVE': '#inclusive',
    'INCREMENT': '#increment',
    'INCREMENTAL': '#incremental',
    'INDEX': '#index',
    'INDEXED': '#indexed',
    'INDEXES': '#indexes',
    'INDICATOR': '#indicator',
    'INFINITE': '#infinite',
    'INITIALLY': '#initially',
    'INLINE': '#inline',
    'INNER': '#inner',
    'INNTER': '#innter',
    'INOUT': '#inout',
    'INPUT': '#input',
    'INSENSITIVE': '#insensitive',
    'INSERT': '#insert',
    'INSTEAD': '#instead',
    'INT': '#int',
    'INTEGER': '#integer',
    'INTERSECT': '#intersect',
    'INTERVAL': '#interval',
    'INTO': '#into',
    'INVALIDATE': '#invalidate',
    'IS': '#is',
    'ISOLATION': '#isolation',
    'ITEM': '#item',
    'ITEMS': '#items',
    'ITERATE': '#iterate',
    'JOIN': '#join',
    'KEY': '#key',
    'KEYS': '#keys',
    'LAG': '#lag',
    'LANGUAGE': '#language',
    'LARGE': '#large',
    'LAST': '#last',
    'LATERAL': '#lateral',
    'LEAD': '#lead',
    'LEADING': '#leading',
    'LEAVE': '#leave',
    'LEFT': '#left',
    'LENGTH': '#length',
    'LESS': '#less',
    'LEVEL': '#level',
    'LIKE': '#like',
    'LIMIT': '#limit',
    'LIMITED': '#limited',
    'LINES': '#lines',
    'LIST': '#list',
    'LOAD': '#load',
    'LOCAL': '#local',
    'LOCALTIME': '#localtime',
    'LOCALTIMESTAMP': '#localtimestamp',
    'LOCATION': '#location',
    'LOCATOR': '#locator',
    'LOCK': '#lock',
    'LOCKS': '#locks',
    'LOG': '#log',
    'LOGED': '#loged',
    'LONG': '#long',
    'LOOP': '#loop',
    'LOWER': '#lower',
    'MAP': '#map',
    'MATCH': '#match',
    'MATERIALIZED': '#materialized',
    'MAX': '#max',
    'MAXLEN': '#maxlen',
    'MEMBER': '#member',
    'MERGE': '#merge',
    'METHOD': '#method',
    'METRICS': '#metrics',
    'MIN': '#min',
    'MINUS': '#minus',
    'MINUTE': '#minute',
    'MISSING': '#missing',
    'MOD': '#mod',
    'MODE': '#mode',
    'MODIFIES': '#modifies',
    'MODIFY': '#modify',
    'MODULE': '#module',
    'MONTH': '#month',
    'MULTI': '#multi',
    'MULTISET': '#multiset',
    'NAME': '#name',
    'NAMES': '#names',
    'NATIONAL': '#national',
    'NATURAL': '#natural',
    'NCHAR': '#nchar',
    'NCLOB': '#nclob',
    'NEW': '#new',
    'NEXT': '#next',
    'NO': '#no',
    'NONE': '#none',
    'NOT': '#not',
    'NULL': '#null',
    'NULLIF': '#nullif',
    'NUMBER': '#number',
    'NUMERIC': '#numeric',
    'OBJECT': '#object',
    'OF': '#of',
    'OFFLINE': '#offline',
    'OFFSET': '#offset',
    'OLD': '#old',
    'ON': '#on',
    'ONLINE': '#online',
    'ONLY': '#only',
    'OPAQUE': '#opaque',
    'OPEN': '#open',
    'OPERATOR': '#operator',
    'OPTION': '#option',
    'OR': '#or',
    'ORDER': '#order',
    'ORDINALITY': '#ordinality',
    'OTHER': '#other',
    'OTHERS': '#others',
    'OUT': '#out',
    'OUTER': '#outer',
    'OUTPUT': '#output',
    'OVER': '#over',
    'OVERLAPS': '#overlaps',
    'OVERRIDE': '#override',
    'OWNER': '#owner',
    'PAD': '#pad',
    'PARALLEL': '#parallel',
    'PARAMETER': '#parameter',
    'PARAMETERS': '#parameters',
    'PARTIAL': '#partial',
    'PARTITION': '#partition',
    'PARTITIONED': '#partitioned',
    'PARTITIONS': '#partitions',
    'PATH': '#path',
    'PERCENT': '#percent',
    'PERCENTILE': '#percentile',
    'PERMISSION': '#permission',
    'PERMISSIONS': '#permissions',
    'PIPE': '#pipe',
    'PIPELINED': '#pipelined',
    'PLAN': '#plan',
    'POOL': '#pool',
    'POSITION': '#position',
    'PRECISION': '#precision',
    'PREPARE': '#prepare',
    'PRESERVE': '#preserve',
    'PRIMARY': '#primary',
    'PRIOR': '#prior',
    'PRIVATE': '#private',
    'PRIVILEGES': '#privileges',
    'PROCEDURE': '#procedure',
    'PROCESSED': '#processed',
    'PROJECT': '#project',
    'PROJECTION': '#projection',
    'PROPERTY': '#property',
    'PROVISIONING': '#provisioning',
    'PUBLIC': '#public',
    'PUT': '#put',
    'QUERY': '#query',
    'QUIT': '#quit',
    'QUORUM': '#quorum',
    'RAISE': '#raise',
    'RANDOM': '#random',
    'RANGE': '#range',
    'RANK': '#rank',
    'RAW': '#raw',
    'READ': '#read',
    'READS': '#reads',
    'REAL': '#real',
    'REBUILD': '#rebuild',
    'RECORD': '#record',
    'RECURSIVE': '#recursive',
    'REDUCE': '#reduce',
    'REF': '#ref',
    'REFERENCE': '#reference',
    'REFERENCES': '#references',
    'REFERENCING': '#referencing',
    'REGEXP': '#regexp',
    'REGION': '#region',
    'REINDEX': '#reindex',
    'RELATIVE': '#relative',
    'RELEASE': '#release',
    'REMAINDER': '#remainder',
    'RENAME': '#rename',
    'REPEAT': '#repeat',
    'REPLACE': '#replace',
    'REQUEST': '#request',
    'RESET': '#reset',
    'RESIGNAL': '#resignal',
    'RESOURCE': '#resource',
    'RESPONSE': '#response',
    'RESTORE': '#restore',
    'RESTRICT': '#restrict',
    'RESULT': '#result',
    'RETURN': '#return',
    'RETURNING': '#returning',
    'RETURNS': '#returns',
    'REVERSE': '#reverse',
    'REVOKE': '#revoke',
    'RIGHT': '#right',
    'ROLE': '#role',
    'ROLES': '#roles',
    'ROLLBACK': '#rollback',
    'ROLLUP': '#rollup',
    'ROUTINE': '#routine',
    'ROW': '#row',
    'ROWS': '#rows',
    'RULE': '#rule',
    'RULES': '#rules',
    'SAMPLE': '#sample',
    'SATISFIES': '#satisfies',
    'SAVE': '#save',
    'SAVEPOINT': '#savepoint',
    'SCAN': '#scan',
    'SCHEMA': '#schema',
    'SCOPE': '#scope',
    'SCROLL': '#scroll',
    'SEARCH': '#search',
    'SECOND': '#second',
    'SECTION': '#section',
    'SEGMENT': '#segment',
    'SEGMENTS': '#segments',
    'SELECT': '#select',
    'SELF': '#self',
    'SEMI': '#semi',
    'SENSITIVE': '#sensitive',
    'SEPARATE': '#separate',
    'SEQUENCE': '#sequence',
    'SERIALIZABLE': '#serializable',
    'SESSION': '#session',
    'SET': '#set',
    'SETS': '#sets',
    'SHARD': '#shard',
    'SHARE': '#share',
    'SHARED': '#shared',
    'SHORT': '#short',
    'SHOW': '#show',
    'SIGNAL': '#signal',
    'SIMILAR': '#similar',
    'SIZE': '#size',
    'SKEWED': '#skewed',
    'SMALLINT': '#smallint',
    'SNAPSHOT': '#snapshot',
    'SOME': '#some',
    'SOURCE': '#source',
    'SPACE': '#space',
    'SPACES': '#spaces',
    'SPARSE': '#sparse',
    'SPECIFIC': '#specific',
    'SPECIFICTYPE': '#specifictype',
    'SPLIT': '#split',
    'SQL': '#sql',
    'SQLCODE': '#sqlcode',
    'SQLERROR': '#sqlerror',
    'SQLEXCEPTION': '#sqlexception',
    'SQLSTATE': '#sqlstate',
    'SQLWARNING': '#sqlwarning',
    'START': '#start',
    'STATE': '#state',
    'STATIC': '#static',
    'STATUS': '#status',
    'STORAGE': '#storage',
    'STORE': '#store',
    'STORED': '#stored',
    'STREAM': '#stream',
    'STRING': '#string',
    'STRUCT': '#struct',
    'STYLE': '#style',
    'SUB': '#sub',
    'SUBMULTISET': '#submultiset',
    'SUBPARTITION': '#subpartition',
    'SUBSTRING': '#substring',
    'SUBTYPE': '#subtype',
    'SUM': '#sum',
    'SUPER': '#super',
    'SYMMETRIC': '#symmetric',
    'SYNONYM': '#synonym',
    'SYSTEM': '#system',
    'TABLE': '#table',
    'TABLESAMPLE': '#tablesample',
    'TEMP': '#temp',
    'TEMPORARY': '#temporary',
    'TERMINATED': '#terminated',
    'TEXT': '#text',
    'THAN': '#than',
    'THEN': '#then',
    'THROUGHPUT': '#throughput',
    'TIME': '#time',
    'TIMESTAMP': '#timestamp',
    'TIMEZONE': '#timezone',
    'TINYINT': '#tinyint',
    'TO': '#to',
    'TOKEN': '#token',
    'TOTAL': '#total',
    'TOUCH': '#touch',
    'TRAILING': '#trailing',
    'TRANSACTION': '#transaction',
    'TRANSFORM': '#transform',
    'TRANSLATE': '#translate',
    'TRANSLATION': '#translation',
    'TREAT': '#treat',
    'TRIGGER': '#trigger',
    'TRIM': '#trim',
    'TRUE': '#true',
    'TRUNCATE': '#truncate',
    'TTL': '#ttl',
    'TUPLE': '#tuple',
    'TYPE': '#type',
    'UNDER': '#under',
    'UNDO': '#undo',
    'UNION': '#union',
    'UNIQUE': '#unique',
    'UNIT': '#unit',
    'UNKNOWN': '#unknown',
    'UNLOGGED': '#unlogged',
    'UNNEST': '#unnest',
    'UNPROCESSED': '#unprocessed',
    'UNSIGNED': '#unsigned',
    'UNTIL': '#until',
    'UPDATE': '#update',
    'UPPER': '#upper',
    'URL': '#url',
    'USAGE': '#usage',
    'USE': '#use',
    'USER': '#user',
    'USERS': '#users',
    'USING': '#using',
    'UUID': '#uuid',
    'VACUUM': '#vacuum',
    'VALUE': '#value',
    'VALUED': '#valued',
    'VALUES': '#values',
    'VARCHAR': '#varchar',
    'VARIABLE': '#variable',
    'VARIANCE': '#variance',
    'VARINT': '#varint',
    'VARYING': '#varying',
    'VIEW': '#view',
    'VIEWS': '#views',
    'VIRTUAL': '#virtual',
    'VOID': '#void',
    'WAIT': '#wait',
    'WHEN': '#when',
    'WHENEVER': '#whenever',
    'WHERE': '#where',
    'WHILE': '#while',
    'WINDOW': '#window',
    'WITH': '#with',
    'WITHIN': '#within',
    'WITHOUT': '#without',
    'WORK': '#work',
    'WRAPPED': '#wrapped',
    'WRITE': '#write',
    'YEAR': '#year',
    'ZONE': '#zone'
}




from dynamof.core import builder as ab
from dynamof.core import args
from dynamof.core.utils import shake
from dynamof.core.model import Operation, Condition
from dynamof.core.response import response


def query(table_name: str, conditions: Condition, index_name: str = None):
    """Creates an Operation that will query a table for items that match the given
    conditions when run.

    Args:
        table_name (str): The name of the table to query
        conditions (Condition): A condition object that should be used to determine which items to return
            Examples:
                - query('products', conditions: attr('title').equals('Old Buggy'))
                - query('products', conditions: cand(attr('type').equals('car'), attr('color').equals('red'))

    """
    build = ab.builder(
        table_name=table_name,
        index_name=index_name,
        conditions=conditions)
    description = shake(
        TableName=build(args.TableName),
        IndexName=build(args.IndexName),
        KeyConditionExpression=build(args.KeyConditionExpression),
        ExpressionAttributeNames=build(args.ExpressionAttributeNames),
        ExpressionAttributeValues=build(args.ExpressionAttributeValues))
    return Operation(description, run)

def run(client, description):
    res = client.query(**description)
    return response(res)



from boto3.dynamodb.types import TypeSerializer

from dynamof.core.utils import guid, merge, update, pipe, find
from dynamof.core.model import AttributeGroup, Attribute, RequestTree
from dynamof.core.dynamo import get_safe_alias
from dynamof.core.Immutable import Immutable


def parse_attr(key, value):

    def make_attr(k, v):
        return Attribute(k, k, v, k, None)

    def replace_reserved_key(attr):
        """Finds reserved dynamo keywords in the attribute
        names and sets an alias the is safe to use instead"""
        return update(attr, alias=get_safe_alias(attr.original))

    def build_key(attr):
        """Builds the key that can be used to reference
        the attribute's value in an expression"""
        return update(attr, key=f':{attr.key}')

    def apply_function_values(attr):
        """Allows us to support passing a special Function as the
        value of the attribute. Here, if the value property is a
        Function we move it to the func property to be used later
        and call it to get the real value"""
        is_function = lambda val: isinstance(val, Immutable) and val.expression is not None and val.value is not None
        if is_function(attr.value):
            fn = attr.value
            return update(attr,
                func=fn,
                value=fn.value())
        return attr

    def build_value_type_tree(attr):
        """Last step in parsing pipeline, reads the value of the
        attribute and uses boto's serializer to convert it to that
        freaky tree with type indicators that dynamo requires"""
        return update(attr,
            value=TypeSerializer().serialize(attr.value))

    @pipe(make_attr)
    @pipe(replace_reserved_key)
    @pipe(build_key)
    @pipe(apply_function_values)
    @pipe(build_value_type_tree)
    def parse(a):
        return a

    return parse(key, value)

def parse_key(key_name):

    if key_name is None:
        return None

    annotation = key_name[-4:]
    key_type = 'N' if annotation == ':int' else 'S'
    key_name = key_name[:-4] if annotation in [':int', ':str'] else key_name

    return {
        'name': key_name,
        'type': key_type
    }


def builder(
    table_name,
    index_name=None,
    key=None,
    attributes=None,
    conditions=None,
    hash_key=None,
    auto_id=None,
    range_key=None,
    gsi=None,
    lsi=None):

    key = key or {}
    attributes = attributes or {}
    condition_attrs = conditions.attributes if conditions is not None else {}

    if auto_id is not None:
        attributes = {
            **attributes,
            auto_id: guid()
        }

    attrs = AttributeGroup(
        keys=[parse_attr(k, v) for k, v in key.items()],
        values=[parse_attr(k, v) for k, v in attributes.items()],
        conditions=[parse_attr(k, v) for k, v in condition_attrs.items()])

    def update_dicts_in_list(list_of_dicts, keys=None, with_fn=None):
        if list_of_dicts is None or keys is None or with_fn is None:
            return list_of_dicts
        return [{
            k: v if k not in keys else with_fn(v) for k, v in obj.items()
        } for obj in list_of_dicts]

    hash_key = parse_key(hash_key)
    range_key = parse_key(range_key)
    gsi = update_dicts_in_list(gsi, keys=['range_key', 'hash_key'], with_fn=parse_key)
    lsi = update_dicts_in_list(lsi, keys=['range_key'], with_fn=parse_key)

    return lambda fn: fn(RequestTree(attrs, table_name, index_name, hash_key, range_key, conditions, gsi, lsi))



from dynamof.operations.add import add
from dynamof.operations.create import create
from dynamof.operations.delete import delete
from dynamof.operations.describe import describe
from dynamof.operations.find import find
from dynamof.operations.query import query
from dynamof.operations.update import update
from dynamof.operations.scan import scan

from dynamof.core.utils import immutable


def table(db, table_name):

    def wrap_op(op):
        def call_op(*args, **kwargs):
            return db(op(table_name, *args, **kwargs))
        return call_op

    return immutable(
        add=wrap_op(add),
        create=wrap_op(create),
        delete=wrap_op(delete),
        describe=wrap_op(describe),
        find=wrap_op(find),
        query=wrap_op(query),
        update=wrap_op(update),
        scan=wrap_op(scan))




from dynamof.core import builder as ab
from dynamof.core import args
from dynamof.core.utils import shake
from dynamof.core.model import Operation
from dynamof.core.response import response


def scan(table_name: str):
    """Creates an Operation that will scan through every item in a table when run.

    Args:
        table_name (str): The name of the table to scan

    """
    build = ab.builder(
        table_name=table_name)
    description = shake(
        TableName=build(args.TableName))
    return Operation(description, run)

def run(client, description):
    res = client.scan(**description)
    return response(res)



import json
import logging

from botocore.exceptions import ClientError

from dynamof.core.exceptions import factory


logger = logging.getLogger('dynamof')

def execute(client, operation):

    runner = operation.runner
    description = operation.description

    logger.debug('Executing Operation:')
    logger.debug(json.dumps(description, indent=2))

    try:
        return runner(client, description)
    except ClientError as err:
        raise factory(err)




from dynamof.core.utils import immutable


def Attr(equals, gt, lt, lt_or_eq, gt_or_eq, between, begins_with):
    return immutable({
        'equals': equals,
        'gt': gt,
        'lt': lt,
        'lt_or_eq': lt_or_eq,
        'gt_or_eq': gt_or_eq,
        'between': between,
        'begins_with': begins_with
    })

def Operation(description, runner):
    return immutable({
        'description': description,
        'runner': runner
    })

def Condition(expression, attributes, references=None):
    references = references or []
    return immutable({
        'expression': expression,
        'attributes': attributes,
        'references': references
    })

def Function(expression, value):
    return immutable({
        'expression': expression,
        'value': value
    })

def RequestTree(attributes, table_name, index_name, hash_key, range_key, conditions, gsi, lsi):
    return immutable({
        'attributes': attributes,
        'table_name': table_name,
        'index_name': index_name,
        'hash_key': hash_key,
        'range_key': range_key,
        'conditions': conditions,
        'gsi': gsi,
        'lsi': lsi
    })

def AttributeGroup(keys, values, conditions):
    return immutable({
        'keys': keys,
        'values': values,
        'conditions': conditions
    })

def Attribute(original, key, value, alias, func):
    """
    Parameters
    ----------
    original : str
        The orignial name the client passed us for the attribute.
        Should never change for later references

        Ex. { 'item': 'a' } => 'item'


    key : str
        The key we should use in expressions to refer to the value
        of this attribute

        Ex. { 'item': 'a' } => ':item'

    value : dict
        The typed value of the attribute. NOTE: The attribute name should
        be stripped out so only the object defining the value/type is stored
        here. Different operations will use this and some will need to set a
        custom key so here were only storing the value of the value/type tree

        Ex. { 'item': 'a' } => { 'S': 'a' }

    alias : str
        The name we should use when refering to the dynamodb _column_. Typically
        this will be the same as `original`. However, in the case of reserved attr
        names it will be updated to something dynamodb friendly

        Ex. { 'item': 'a' } => '#item'

    func : Function
        A func from `dynamof.funcs` that should be called to modify the attr
    """
    return immutable({
        'original': original,
        'key': key,
        'value': value,
        'alias': alias,
        'func': func
    })



import uuid
import json
import decimal
import functools
from collections import ChainMap

from dynamof.core.Immutable import Immutable


def guid():
    return str(uuid.uuid4())

def shake(**kwargs):
    """Removes none and empty object values"""
    return { k: v for k, v in kwargs.items() if bool(v) and v is not None }

def merge(list_of_dicts):
    return dict(ChainMap(*list_of_dicts))

def flatten(list_of_lists):
    return [item for sublist in list_of_lists for item in sublist]

def strip_Decimals(obj):

    if isinstance(obj, list):
        return [strip_Decimals(item) for item in obj]
    if isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = strip_Decimals(v)
        return obj
    if isinstance(obj, decimal.Decimal):
        if obj % 1 == 0:
            return int(obj)
        return float(obj)
    return obj

def find(a_list, fn, default=None):
    return next((item for item in a_list if fn(item)), default)

def update(obj, **kwargs):
    """Returns a new instance of the given object with
    all key/val in kwargs set on it
    """
    return immutable({
        **obj,
        **kwargs
    })

def immutable(obj=None, **kwargs):
    if obj is not None:
        return Immutable(**obj)
    return Immutable(**kwargs)

def pipe(transform):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(transform(*args, **kwargs))
        return wrapper
    return decorator




from dynamof.core import builder as ab
from dynamof.core import args
from dynamof.core.utils import shake
from dynamof.core.model import Operation
from dynamof.core.response import response


def create(table_name: str, hash_key: str, range_key: str = None, gsi: str = None, lsi: str = None):
    """Creates an Operation that will create a table when run.

    Args:
        table_name (str): The name of the table to create
        hash_key (str): The name of the hash key to use for the table
        range_key (str, optional): If provided, will be used as the name of the range key attribute for the table
        gsi (str, optional): If provided, will be used as the name for a global secondary index
        lsi (str, optional): If provided, will be used as the name for a local secondary index

    """
    build = ab.builder(
        table_name=table_name,
        hash_key=hash_key,
        range_key=range_key,
        gsi=gsi,
        lsi=lsi)
    description = shake(
        TableName=build(args.TableName),
        KeySchema=build(args.KeySchema),
        AttributeDefinitions=build(args.AttributeDefinitions),
        ProvisionedThroughput=build(args.ProvisionedThroughput),
        LocalSecondaryIndexes=build(args.LocalSecondaryIndexes),
        GlobalSecondaryIndexes=build(args.GlobalSecondaryIndexes))
    return Operation(description, run)

def run(client, description):
    res = client.create_table(**description)
    return response(res)



import setuptools

VERSION = '2.1.8'

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='dynamof',
    version=VERSION,
    author='Ray Epps',
    author_email='rayharryepps@gmail.com',
    description='A small interface for more easily making calls to dynamo using boto.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/rayepps/dynamof',
    packages=[
        'dynamof',
        'dynamof.core',
        'dynamof.operations'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3',
    install_requires=[
        'boto3',
        'botocore'
    ]
)




from dynamof.core import builder as ab
from dynamof.core import args
from dynamof.core.utils import shake
from dynamof.core.model import Operation, Condition
from dynamof.core.response import response


def update(table_name: str, key: dict, attributes: dict, conditions: Condition = None):
    """Creates an Operation that will update any items that match the given key with the given
    attributes when run.

    Args:
        table_name (str): The name of the table to update items in
        key (dict): The key that represents an identifier for the item/s to update
            Examples:
                - { 'username': 'sunshie' }
                - { 'id': '9snw82h', 'type': 'car' } -> when table uses hash key and range key
        attributes (dict): An arbitrary object of key values to set on the item when updating
            Examples:
                - { 'color': 'red', 'size': 'large' }
                - { 'username': 'sunshie12' }
        conditions (Condition, optional): A condition object that should be used to filter which items are updated
            Examples:
                - query('products', conditions: attr('title').equals('Old Buggy'))
                - query('products', conditions: cand(attr('type').equals('car'), attr('color').equals('red'))

    """
    build = ab.builder(
        table_name=table_name,
        key=key,
        attributes=attributes,
        conditions=conditions)
    description = shake(
        TableName=build(args.TableName),
        Key=build(args.Key),
        ConditionExpression=build(args.ConditionExpression),
        UpdateExpression=build(args.UpdateExpression),
        ExpressionAttributeNames=build(args.ExpressionAttributeNames),
        ExpressionAttributeValues=build(args.ExpressionAttributeValues),
        ReturnValues='ALL_NEW')
    return Operation(description, run)

def run(client, description):
    res = client.update_item(**description)
    return response(res)




from dynamof.core import builder as ab
from dynamof.core import args
from dynamof.core.utils import shake
from dynamof.core.model import Operation, Condition
from dynamof.core.response import response


def delete(table_name: str, key: dict, conditions: Condition = None):
    """Creates an Operation that will create a table when run.

    Args:
        table_name (str): The name of the table to delete an item from
        key (dict): An object with a key, value that will be used as the identifier of the item to delete
            Examples:
                - { 'username': 'sunshie' }
                - { 'id': '9snw82h', 'type': 'car' } -> when table uses hash key and range key
        conditions (Condition, optional): If provided, will be used to create a condition for the delete
            Example:
                - delete('products', { 'id': 12 }, conditions: attr('title').equals('Old Buggy'))

    """
    build = ab.builder(
        table_name=table_name,
        key=key,
        conditions=conditions)
    description = shake(
        TableName=build(args.TableName),
        Key=build(args.Key),
        ConditionExpression=build(args.ConditionExpression),
        ExpressionAttributeValues=build(args.ExpressionAttributeValues))
    return Operation(description, run)

def run(client, description):
    res = client.delete_item(**description)
    return response(res)




from boto3.dynamodb.types import TypeDeserializer

from dynamof.core.utils import strip_Decimals, immutable


def destructure_type_tree(data):

    if data is None:
        return None

    # Using `strip_Decimals` here to patch an
    # undesireable behavior with dynamo where
    # it takes in number types but always returns
    # Decimal class type.
    # https://github.com/boto/boto3/issues/369

    return strip_Decimals({
        k: TypeDeserializer().deserialize(v) for k, v in data.items()
    })

def response(res):

    def get_items():
        items = res.get('Items', None)
        if items is not None and len(items) > 0:
            return [destructure_type_tree(item) for item in items]
        return None

    return immutable({

        'retries': lambda: res.get('ResponseMetadata', {}).get('RetryAttempts', None),

        'success': lambda: res.get('ResponseMetadata', {}).get('HTTPStatusCode', 0) == 200,

        'item': lambda: destructure_type_tree(res.get('Item', None)),

        'items': get_items,

        'count': lambda: res.get('Count', None),

        'scanned_count': lambda: res.get('ScannedCount', None),

        'raw': lambda: res

    })



"""Args is a collection of function that take in a request and return a specific
boto3 argument"""

from dynamof.core.model import RequestTree
from dynamof.core.utils import merge, immutable
from dynamof.core.dynamo import get_safe_alias


def TableName(request: RequestTree):
    return request.table_name

def Key(request: RequestTree):
    """Creates the `Key` argument for a boto3 request description.
    @param request: ResultTree
    @return dict

    Example:

    return { 'id': { 'S': 'ab384020' }}
    """
    return merge([{
        key.alias: key.value
    } for key in request.attributes.keys])

def ConditionExpression(request: RequestTree):
    """Creates the `ConditionExpression` argument for a boto3 request description.
    @param request: ResultTree
    @return dict

    Example:

    return "Price > :limit"
    """
    if request.conditions is None:
        return None
    return request.conditions.expression(request.attributes.conditions)

def KeyConditionExpression(request: RequestTree):
    if request.conditions is None:
        return None
    return request.conditions.expression(request.attributes.conditions)

def UpdateExpression(request: RequestTree):
    def expression(attr):
        if attr.func is not None:
            return attr.func.expression(attr)
        return f'{attr.alias} = {attr.key}'
    key_expressions = [expression(key) for key in request.attributes.values]
    key_expression = ', '.join(key_expressions)
    return f'SET {key_expression}'

def ExpressionAttributeNames(request: RequestTree):
    all_attributes = [
        *request.attributes.keys,
        *request.attributes.values,
        *request.attributes.conditions
    ]

    aliased_attributes = [attr for attr in all_attributes if attr.alias[0] == '#']

    if request.conditions is not None and request.conditions.references is not None:
        for ref in request.conditions.references:
            aliased_attributes.append(immutable({
                'alias': get_safe_alias(ref),
                'original': ref
            }))

    return { attr.alias: attr.original for attr in aliased_attributes }

def Item(request: RequestTree):
    return merge([
        { attr.original: attr.value } for attr in request.attributes.values
    ])

def ExpressionAttributeValues(request: RequestTree):
    all_attributes = [
        *request.attributes.values,
        *request.attributes.conditions
    ]
    return {
        attr.key: attr.value for attr in all_attributes
    }

def KeySchema(request: RequestTree):
    schema = [
        {
            'AttributeName': request.hash_key.get('name'),
            'KeyType': 'HASH'
        }
    ]
    if request.range_key is not None:
        schema.append({
            'AttributeName': request.range_key.get('name'),
            'KeyType': 'RANGE'
        })
    return schema


def AttributeDefinitions(request: RequestTree):
    '''Finds all the hash keys and range keys in the given
    request (including indexes) and sets them in the boto
    standard AttributeDefinitions argument model. Also, looks
    for type annotations in the key names ('key_name:str' or 'key_name:int'),
    strips them from the name, and uses them to set the AttributeType.'''

    remove_duplicates = lambda list_of_keys: [dict(t) for t in {tuple(d.items()) for d in list_of_keys}]
    remove_nones = lambda list_of_keys: [k for k in list_of_keys if k is not None]

    all_keys = remove_duplicates(remove_nones([
        request.hash_key,
        request.range_key,
        *[i.get('range_key') for i in request.gsi or []],
        *[i.get('hash_key') for i in request.gsi or []],
        *[i.get('range_key') for i in request.lsi or []]
    ]))

    return [{
        'AttributeName': key.get('name'),
        'AttributeType': key.get('type')
    } for key in all_keys]

def ProvisionedThroughput(request: RequestTree): # pylint: disable=unused-argument
    return {
        'ReadCapacityUnits': 1,
        'WriteCapacityUnits': 1
    }

def LocalSecondaryIndexes(request: RequestTree):

    def make_index(lsi):
        name = lsi.get('name')
        range_key = lsi.get('range_key').get('name')
        hash_key = request.hash_key.get('name')
        return {
            'IndexName': name,
            'KeySchema': [
                {
                    'AttributeName': hash_key,
                    'KeyType': 'HASH'
                },
                {
                    'AttributeName': range_key,
                    'KeyType': 'RANGE'
                }
            ],
            'Projection': {
                'ProjectionType': 'ALL'
            }
        }

    if request.lsi is None:
        return None

    return [
        make_index(i) for i in request.lsi
    ]

def GlobalSecondaryIndexes(request: RequestTree):

    def make_index(gsi):
        name = gsi.get('name')
        hash_key = gsi.get('hash_key').get('name')
        range_key = gsi.get('range_key', {}).get('name', None)
        throughput = gsi.get('throughput', 10)
        return {
            'IndexName': name,
            'KeySchema': [
                {
                    'AttributeName': name,
                    'KeyType': type
                } for name, type in [[hash_key, 'HASH'], [range_key, 'RANGE']] if name is not None
            ],
            'Projection': {
                'ProjectionType': 'ALL'
            },
            'ProvisionedThroughput': {
                'ReadCapacityUnits': throughput,
                'WriteCapacityUnits': throughput
            }
        }

    if request.gsi is None:
        return None

    return [
        make_index(i) for i in request.gsi
    ]

def IndexName(request: RequestTree):
    return request.index_name
